#!/usr/bin/env python
# coding: utf-8

# # Evaluating Design Performance
# 
# In this example, we assume you have encoded an NLoed model and generated an optimal design. Before implementing the design you may wish to get a better and more interpretable understanding of its expected performance.
# 
# To begin we import the results for the *N2_designoptimization* notebook:

# In[ ]:


from scripts.N2_design_optimization import *

